#!/usr/bin/ruby
#
# DNS Balance --- ưŪʬԤʤ DNS 
#
# By: YOKOTA Hiroshi <yokota@netlab.is.tsukuba.ac.jp>

# $Id: dns_balance.rb,v 1.22 2002/09/26 11:45:01 elca Exp $

# DNS Balance ¸ߤѥ̾
$prefix = "/usr/local/share/DNS-Balance-1.0.2"
$LOAD_PATH.unshift($prefix)

require 'socket'
require 'thread'
require 'getopts'

require 'datatype.rb'
require 'multilog.rb'
require 'log_writer.rb'
require 'util.rb'
require 'as_search.rb'

require 'namespace.rb'
require 'addrdb.rb'

#####################################################################
# 桼㳰
class DnsNotImplementedError < StandardError ; end
class DnsTruncatedError      < StandardError ; end
class DnsNoQueryError        < StandardError ; end
class DnsNoMoreResourceError < StandardError ; end
#
class DnsNsRequestError      < StandardError ; end
class DnsSoaRequestError     < StandardError ; end


###################################################################
# ؿ

def usage()
  print "Usage: ruby dns_balance.rb [-lh] [-i ipaddr]\n"
  print "       -i ipaddr  listen IP address (default:0.0.0.0)\n"
  print "       -d nameserver:ipaddr\n"
  print "                  specify my domain name and this nameserver's addr\n"
  print "       -h         help message\n"
  print "       --as       Enable AS namespace\n"
  exit(111)
end

# DNS ѥåȤƤȼΥפȼΥ饹Ф
def parse_packet(packet)
  (number, flags, num_q, ans_rr, ort_rr, add_rr, str) =  packet.unpack("a2 a2 a2 a2 a2 a2 a*")

  if num_q != "\0\1"
    return ["\0", "\0\0", "\0\0"]
  end

  # ̯ʥѥåȤӽ
  #  1 ʤ餳ΩĤϤ
  if (str.split("\0")[0].length+1 +2+2) != str.length
    return ["\0", "\0\0", "\0\0"]
  end

  (q, q_type, q_class) = str.unpack("a#{(str.length-4).to_s} a2 a2")

  return [q, q_type, q_class]
end

# 饤ȤIPɥ쥹֤
def parse_client_addr(str)
  (family, port, ipaddr, pad) = str.unpack("a2 a2 a4 a*")

  return ipaddr
end

# 饤ȤIPɥ쥹ˤäƤѤ
# ֤̾ƤϤޤʪʤ "default" ˤʤ
def select_namespace(addrstr, name)

  netaddrs = []
  netaddrs << ipstr_to_str(addrstr)
  (0..31).each {
    |i|
    mask = 32-i

    netaddrs << (ipstr_to_str(ipstr_mask(addrstr, mask)) + "/" + mask.to_s)
  }

  # custom namespace
  netaddrs.each {
    |i|
    if $namespace_db[i]                  != nil &&
	$addr_db[$namespace_db[i]]       != nil &&
	$addr_db[$namespace_db[i]][name] != nil
      return $namespace_db[i]
    end
  }

  # address number namespace
  netaddrs.each {
    |i|
    if $addr_db[i]        != nil &&
	$addr_db[i][name] != nil
      return i
    end
  }

  # AS namespace
  if $OPT_as &&
      # RFC1918 / ץ饤١ȥɥ쥹Ϥɤ AS ˤ°Ƥʤ
      ipstr_to_str(ipstr_mask(addrstr,  8)) != "10.0.0.0"      &&
      ipstr_to_str(ipstr_mask(addrstr, 12)) != "172.16.0.0"    &&
      ipstr_to_str(ipstr_mask(addrstr, 16)) != "192.168.0.0"   &&
      ipstr_to_str(ipstr_mask(addrstr, 21)) != "204.152.184.0" &&
      ipstr_to_str(addrstr                ) != "127.0.0.1"

    tmp = ipstr_to_str(addrstr)
    as = as_search(tmp)
    if as                  != nil &&
	$addr_db[as]       != nil &&
	$addr_db[as][name] != nil
      return as
    end
  end

  return "default"
end

# ŤߤĤѿΤɽ
def make_rand_array(namespace, name)
  rnd_max = 0
  rnd_slesh = []

  $addr_db[namespace][name].each {
    |i|
    rnd_max += (10000 - min(10000, i[1])) # badness κͤ 10000
    rnd_slesh.push(rnd_max)
  }

  return [rnd_max, rnd_slesh]
end

# ŤߤĤ
def select_rand_array(namespace, name, size)
  (rnd_max, rnd_slesh) = make_rand_array(namespace, name)

  if rnd_max == 0  # ƤΥۥȤ Badness  10000 ä
    return []
  end

  arr = []
  (0...size).each {
    |i|
    rnd = rand(rnd_max)
    (0...rnd_slesh.size).each {
      |j|
      if rnd <= rnd_slesh[j]
	arr.push(j)
	break
      end
    }
  }

  return arr
end

# ѥåȤå
def check_packet(q, q_type, q_class)
  # ž̵
  if q_type == DnsType::AXFR
    $ml.log("AXFR")
    raise DnsNotImplementedError
  end

  # IP(UDP) Τ߼դ
  if q_class != DnsClass::INET && q_class != DnsClass::ANY
    $ml.log("noIP: " + q_class.dump)
    raise DnsNoQueryError
  end

  # ԲĤʸ
  if (q =~ /[()<>@,;:\\\"\.\[\]]/) != nil
    $ml.log("char")
    raise DnsNoQueryError
  end

  if $nameserver
    if q_type == DnsType::NS
      raise DnsNsRequestError
    end
    if q_type == DnsType::SOA
      raise DnsSoaRequestError
    end
  end

  # A/ANY 쥳ɤΤ߼դ
  if q_type != DnsType::A && q_type != DnsType::ANY
    $ml.log("noA " + q_type.dump)
    raise DnsNoQueryError
  end

end

def make_ns(str)
  (domain, addr) = str.split(":")

  return [str_to_dnsstr(domain), str_to_ipstr(addr)]
end

def add_ns_addr()
  if $addr_db && $addr_db["default"] && \
      $nameserver && $nameserver["domain"]
    $addr_db["default"]["ns." + dnsstr_to_str($nameserver["domain"])] = [[$nameserver["ipstr"].unpack("CCCC"), 0]]
  end
end

def make_soastr()
  dom  = $nameserver["domain"]
  mail = "\x0ahostmaster" + dom
  ttl  = "\0\0\x0e\x10" # 1

  rec = dom + mail +
    [$nameserver["serial"],1800,600,604800,3600].pack("NNNNN")

  return dom + DnsType::SOA + DnsClass::INET + ttl +
    [rec.length].pack("n") + rec
end

def make_nsstr()
  dom = $nameserver["domain"]
  ns  = "\002ns" + dom
  ttl = "\0\0\x0e\x10" # 1

  return [dom + DnsType::NS + DnsClass::INET + ttl + [ns.length].pack("n") + ns,
    ns + DnsType::A + DnsClass::INET + ttl + "\0\4" + $nameserver["ipstr"]]
end



######################################################################
# main

srand()
getopts("lh", "i:0.0.0.0", "d:", "as")

usage() if $OPT_h

$ml = MultiLog.new
$ml.open

if $OPT_d
  (domain, addr) = make_ns($OPT_d)
  $nameserver = {
    "domain" => domain,
    "ipstr"  => addr,
    "serial" => Time.now.to_i,
  }
  # add ns
  add_ns_addr()

  $ml.log("domain: " + dnsstr_to_str(domain))
  $ml.log("server: " + ipstr_to_str(addr))
end

$ml.log("dir: " + $prefix)
$ml.log("start")


#
# ɥ쥹ǡ١ưŪ
#
Thread.start do
  loop do
    if test(?r, $prefix + "/addr") || test(?r, "./addr")
      begin
	load("addr")

	# add ns
	add_ns_addr()

	$ml.log("reload")
      rescue NameError,SyntaxError,Exception
	$ml.log("reload failed")
      end
    end
    #p $addr_db
    sleep(5*60) # 5 ʬ˹
  end
end

gs = Socket.open(Socket::AF_INET, Socket::SOCK_DGRAM, 0)
gs.bind([Socket::AF_INET, 53, str_to_ipstr($OPT_i), ""].pack("s n a4 a8"))

#
# ᥤ롼
#
loop do
  (packet, client_addr) = gs.recvfrom(512)
  Thread.start do
    $SAFE = 2
    begin
      client = parse_client_addr(client_addr)
      (q, q_type, q_class) = parse_packet(packet)
      check_packet(q, q_type, q_class) # -> NoQuery, NotImpl

      name      = dnsstr_to_str(q).downcase
      namespace = select_namespace(client, name)

      if $addr_db[namespace] == nil || $addr_db[namespace][name] == nil
	raise DnsNoQueryError
      end

      if $addr_db[namespace][name].size > 1  # -> NameError -> NoQuery
	size = 8
      else
	size = 1
      end
      a_array = select_rand_array(namespace, name, size)

      if a_array.size == 0
	raise DnsNoMoreResourceError
      end

      # 
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x84   # answer & authenticated
      r[3] &= ~0x0f  # no error

      ans_addrs = []
      a_array.each {
	|i|
	addr = $addr_db[namespace][name][i][0]
	ans_addrs.push(addr)   # ǥХå

	# TTL 򡣶ᤤ꤬ʬäƤʤ TTL Ĺ
	if (a_array.size == 1)
	  ttl = "\0\0\x0e\x10" # 1
	else
	  ttl = "\0\0\0\x3c"   # 60
	end

	#  եåȤ 0x000c
	r += "\xc0\x0c" + DnsType::A + DnsClass::INET + ttl + "\0\4" + addr.pack("CCCC")
      }

      # ο򥻥å
      r[6..7] = [a_array.size].pack("n")

      # ͡ॵФΰ֤򥻥å
      if $nameserver &&
	  Regexp.new($nameserver["domain"] + "$", Regexp::IGNORECASE) =~ q #"
	tmp = make_nsstr()

	r += tmp[0]
	r[8..9] = "\0\1"

	if ("\002ns" + $nameserver["domain"]).downcase != q.downcase
	  r += tmp[1]
	  r[10..11] = "\0\1"
	end
      end

      # Ĺ᤮
      if r.length > 512
	raise DnsTruncatedError
      end

      status = "ok"

    rescue DnsNotImplementedError
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x80  # answer
      r[2] &= ~0x04 # not authenticated
      r[3] &= ~0x0f
      r[3] |= 0x04  # not implemented error
      status    = "NotImpl"

    rescue DnsTruncatedError
      # Ĺ᤮Ϻäƥե饰ΩƤ
      r = r[0,512]
      r[2] |= 0x02
      status = "Truncated"

    rescue DnsNsRequestError
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x84  # answer & authenticated

      if $nameserver &&
	  $nameserver["domain"].downcase == q.downcase

	r[3] &= ~0x0f  # no error

	r += make_nsstr().join("")
	r[6..7]   = "\0\1"
	r[10..11] = "\0\1"

	name      = dnsstr_to_str(q)
	ans_addrs = [$nameserver["ipstr"]]
	status    = "NS"
      else
	r[3] &= ~0x0f
	r[3] |= 0x03  # name error

	name      = dnsstr_to_str(q)
	status    = "NSNG"
      end


    rescue DnsSoaRequestError
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x84  # answer & authenticated
      if $nameserver &&
	  $nameserver["domain"].downcase == q.downcase

	r[3] &= ~0x0f

	r += make_soastr()
	r[6..7] = "\0\1"

	name      = dnsstr_to_str(q)
	ans_addrs = [$nameserver["ipstr"]]
	status    = "SOA"
      else
	r[3] &= ~0x0f
	r[3] |= 0x03  # name error

	name   = dnsstr_to_str(q)
	status = "SOANG"
      end


    rescue DnsNoQueryError,DnsNoMoreResourceError,NameError,StandardError
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x84  # answer & authenticated
      r[3] &= ~0x0f
      r[3] |= 0x03  # name error

      if $nameserver &&
	  Regexp.new($nameserver["domain"] + "$", Regexp::IGNORECASE) =~ q #"

	r += make_soastr()
	r[8..9] = "\0\1"
      end

      status = "NoQuery"

    rescue
      # ˤʤϤ
      r = packet[0,12] + q + q_type + q_class
      r[2] |= 0x80  # answer
      r[2] &= ~0x04 # not authenticated
      r[3] &= ~0x0f
      r[3] |= 0x05  # query refused error
      status = "other"
    end

    #print packet.dump, "\n"
    #print r.dump, "\n"
    #p q

    gs.send(r, 0, client_addr)

    name      = "" if !name
    namespace = "" if !namespace
    ans_addrs = [] if !ans_addrs

    logger($ml, client, status, name, namespace, ans_addrs)

  end
end

# end
